#include "Environment.h"
#include "GvsPreprocessor.h"
#include "GlRenderer.h"
#include "VssRay.h"
#include "ViewCellsManager.h"
#include "Triangle3.h"
#include "IntersectableWrapper.h"
#include "Plane3.h"
#include "RayCaster.h"
#include "Exporter.h"
#include "SamplingStrategy.h"


namespace GtpVisibilityPreprocessor
{
  
struct VizStruct
{
	Polygon3 *enlargedTriangle;
	Triangle3 originalTriangle;
	VssRay *ray;
};

static vector<VizStruct> vizContainer;

GvsPreprocessor::GvsPreprocessor(): 
Preprocessor(), 
//mSamplingType(SamplingStrategy::DIRECTION_BASED_DISTRIBUTION),
mSamplingType(SamplingStrategy::DIRECTION_BOX_BASED_DISTRIBUTION),
mSampleContriPerPass(0), 
mTotalSampleContri(0),
mReverseSamples(0), 
mBorderSamples(0)
{
	Environment::GetSingleton()->GetIntValue("GvsPreprocessor.totalSamples", mTotalSamples);
	Environment::GetSingleton()->GetIntValue("GvsPreprocessor.initialSamples", mInitialSamples);
	Environment::GetSingleton()->GetIntValue("GvsPreprocessor.samplesPerPass", mSamplesPerPass);
	Environment::GetSingleton()->GetFloatValue("GvsPreprocessor.epsilon", mEps);
	Environment::GetSingleton()->GetFloatValue("GvsPreprocessor.threshold", mThreshold);	

	Debug << "Gvs preprocessor options" << endl;
	Debug << "number of total samples: " << mTotalSamples << endl;
	Debug << "number of initial samples: " << mInitialSamples << endl;
	Debug << "number of samples per pass: " << mSamplesPerPass << endl;
	Debug << "threshold: " << mThreshold << endl;
	Debug << "eps: " << mEps << endl;

	mStats.open("gvspreprocessor.log");
}


bool GvsPreprocessor::CheckDiscontinuity(const VssRay &currentRay,
										 const Triangle3 &hitTriangle,
										 const VssRay &oldRay)
{
	const float dist = Magnitude(oldRay.GetDir());
	const float newDist = Magnitude(currentRay.GetDir());
 
#if 0
	if ((dist - newDist) > mThresHold)
#else 
	// rather take relative distance
	if ((dist / newDist) > mThreshold)
#endif
	{
		VssRay *newRay = ReverseSampling(currentRay, hitTriangle, oldRay);
		// ray is not pushed into the queue => can delete ray
		if (!HandleRay(newRay))
			delete newRay;
		
		return true;
	}

	return false;
}


bool GvsPreprocessor::HandleRay(VssRay *vssRay)
{
	const bool storeRaysForViz = true;
	mViewCellsManager->ComputeSampleContribution(*vssRay, true, storeRaysForViz);
		
	// some pvs contribution for this ray?
	if (vssRay->mPvsContribution > 0)
	{
		//cout << " h " << mSampleContriPerPass << " " << mSamplesPerPass << " " << mTotalSamples << endl;
		mRayQueue.push(vssRay);

		if (storeRaysForViz)
		{
			ViewCellContainer::const_iterator vit, vit_end = vssRay->mViewCells.end();
			for (vit = vssRay->mViewCells.begin(); vit != vit_end; ++ vit)
			{
				(*vit)->mVssRays.push_back(new VssRay(*vssRay));
			}
		}
		//mVssRays.push_back(new VssRay(*vssRay));
        ++ mSampleContriPerPass;

		return true;
	}

	return false;
}


/** Creates 3 new vertices for triangle vertex with specified index.
*/
static void CreateNewVertices(VertexContainer &vertices,
							  const Triangle3 &hitTriangle,
							  const VssRay &ray, 
							  const int index,
							  const float eps)
{
	const int indexU = (index + 1) % 3;
	const int indexL = (index == 0) ? 2 : index - 1;

	const Vector3 a = hitTriangle.mVertices[index] - ray.GetOrigin();
	const Vector3 b = hitTriangle.mVertices[indexU] - hitTriangle.mVertices[index];
	const Vector3 c = hitTriangle.mVertices[index] - hitTriangle.mVertices[indexL];
	
	const float len = Magnitude(a);

	const Vector3 dir1 = Normalize(CrossProd(a, b)); //N((pi-xp)(pi+1- pi));
	const Vector3 dir2 = Normalize(CrossProd(a, c)); // N((pi-xp)(pi- pi-1))
	const Vector3 dir3 = DotProd(dir2, dir1) > 0 ? // N((pi-xp)di,i-1+di,i+1(pi-xp))
		Normalize(dir2 + dir1) : Normalize(CrossProd(a, dir1) + CrossProd(dir2, a)); 

	// compute the new three hit points
	// pi, i + 1:  pi+ e|pi-xp|di, j
	const Vector3 pt1 = hitTriangle.mVertices[index] + eps * len * dir1;
	// pi, i - 1:  pi+ e|pi-xp|di, j
    const Vector3 pt2 = hitTriangle.mVertices[index] + eps * len * dir2;
	// pi, i:  pi+ e|pi-xp|di, j
	const Vector3 pt3 = hitTriangle.mVertices[index] + eps * len * dir3;
	
	vertices.push_back(pt2);
	vertices.push_back(pt3);
	vertices.push_back(pt1);
}


void GvsPreprocessor::EnlargeTriangle(VertexContainer &vertices,
									  const Triangle3 &hitTriangle,
									  const VssRay &ray)
{
	CreateNewVertices(vertices, hitTriangle, ray, 0, mEps);
	CreateNewVertices(vertices, hitTriangle, ray, 1, mEps);
	CreateNewVertices(vertices, hitTriangle, ray, 2, mEps);
}


static Vector3 CalcPredictedHitPoint(const VssRay &newRay, 
									 const Triangle3 &hitTriangle,
									 const VssRay &oldRay)
{
	Plane3 plane(hitTriangle.GetNormal(), hitTriangle.mVertices[0]);

	const Vector3 hitPt = 
		plane.FindIntersection(newRay.mTermination, newRay.mOrigin);
	
	return hitPt;
}


static bool EqualVisibility(const VssRay &a, const VssRay &b)
{
	return a.mTerminationObject == b.mTerminationObject;
}


int GvsPreprocessor::SubdivideEdge(const Triangle3 &hitTriangle,
								   const Vector3 &p1, 
								   const Vector3 &p2, 
								   const VssRay &x, 
								   const VssRay &y,
								   const VssRay &oldRay)
{
	// the predicted hitpoint expects to hit the same mesh again
	const Vector3 predictedHitX = CalcPredictedHitPoint(x, hitTriangle, oldRay);
	const Vector3 predictedHitY = CalcPredictedHitPoint(y, hitTriangle, oldRay);

	CheckDiscontinuity(x, hitTriangle, oldRay);
	CheckDiscontinuity(y, hitTriangle, oldRay);

	if (EqualVisibility(x, y))
	{
		return 0;
	}
	else
	{
		cout << "s";
		const Vector3 p = (p1 + p2) * 0.5f;
		SimpleRay sray(oldRay.mOrigin, p - oldRay.mOrigin);
	
		// cast ray into the new subdivision point
		VssRay *newRay = mRayCaster->CastRay(sray, mViewCellsManager->GetViewSpaceBox(), false);
		//cout << "\np1: " << p1 << "\np : " <<  p << "\np2: " << p2 << endl;
		if (!newRay) return 0;

		const bool enqueued = HandleRay(newRay);
		
		// subdivide further
		const int s1 = SubdivideEdge(hitTriangle, p1, p, x, *newRay, oldRay);
		const int s2 = SubdivideEdge(hitTriangle, p, p2, *newRay, y, oldRay);
				
		if (!enqueued)
			delete newRay;
		
		return s1 + s2 + 1;
	}
}


int GvsPreprocessor::AdaptiveBorderSampling(const VssRay &currentRay)
{
	cout << "a";
	Intersectable *tObj = currentRay.mTerminationObject;
	Triangle3 hitTriangle;

	// other types not implemented yet
	if (tObj->Type() == Intersectable::TRIANGLE_INTERSECTABLE)
	{
		hitTriangle = dynamic_cast<TriangleIntersectable *>(tObj)->GetItem();
	}
	else
	{
		cout << "not yet implemented" << endl;
	}

	VertexContainer enlargedTriangle;
	
	/// create 3 new hit points for each vertex
	EnlargeTriangle(enlargedTriangle, hitTriangle, currentRay);
	
	/// create rays from sample points and handle them
	SimpleRayContainer simpleRays;
	simpleRays.reserve(9);

	VertexContainer::const_iterator vit, vit_end = enlargedTriangle.end();

	for (vit = enlargedTriangle.begin(); vit != vit_end; ++ vit)
	{
		const Vector3 rayDir = (*vit) - currentRay.GetOrigin();
		SimpleRay sr(currentRay.GetOrigin(), rayDir);
		simpleRays.AddRay(sr);
	}

	if (0)
	{
		VizStruct dummy;
		dummy.enlargedTriangle = new Polygon3(enlargedTriangle);
		dummy.originalTriangle = hitTriangle;
		//dummy.ray = new VssRay(currentRay);
		vizContainer.push_back(dummy);
	}

	// cast rays to triangle vertices and determine visibility
	VssRayContainer vssRays;
	CastRays(simpleRays, vssRays, false, false);

	// add to ray queue
	EnqueueRays(vssRays);
	
	const int n = (int)enlargedTriangle.size();
	int castRays = (int)vssRays.size();
#if 1
    // recursivly subdivide each edge
	for (int i = 0; i < n; ++ i)
	{
		castRays += SubdivideEdge(
			hitTriangle,
			enlargedTriangle[i], 
			enlargedTriangle[(i + 1) % n], 
			*vssRays[i], 
			*vssRays[(i + 1) % n],
			currentRay);
	}
#endif

	mBorderSamples += castRays;
	return castRays;
}


static Vector3 GetPassingPoint(const VssRay &currentRay,
							   const Triangle3 &hitTriangle,
							   const VssRay &oldRay)
{
	// intersect triangle plane with plane spanned by current samples
	Plane3 plane(currentRay.GetOrigin(), currentRay.GetTermination(), oldRay.GetTermination());
	Plane3 triPlane(hitTriangle.GetNormal(), hitTriangle.mVertices[0]);

	SimpleRay intersectLine = GetPlaneIntersection(plane, triPlane);

	// Evaluate new hitpoint just outside the triangle
	const float factor = 0.95f;
	float t = triPlane.FindT(intersectLine);
	const Vector3 newPoint = intersectLine.mOrigin + t * factor * intersectLine.mDirection;

	return newPoint;
}


VssRay *GvsPreprocessor::ReverseSampling(const VssRay &currentRay,
										 const Triangle3 &hitTriangle,
										 const VssRay &oldRay)
{
	++ mReverseSamples;

	//-- The plane p = (xp, hit(x), hit(xold)) is intersected
	//-- with the newly found triangle (xold is the previous ray from
	//-- which x was generated). On the intersecting line, we select a point
	//-- pnew which lies just outside of the new triangle so the ray
	//-- just passes by inside the gap
    const Vector3 newPoint = GetPassingPoint(currentRay, hitTriangle, oldRay);
	const Vector3 predicted = CalcPredictedHitPoint(currentRay, hitTriangle, oldRay);

	//-- Construct the mutated ray with xnew,dir = predicted(x)- pnew 
	//-- as direction vector
	const Vector3 newDir = predicted - newPoint ;
	// take xnew,p = intersect(viewcell, line(pnew, predicted(x)) as origin ?
	// difficult to say!!
	const Vector3 newOrigin = newDir * -5000.0f;

	return new VssRay(currentRay);
}


int GvsPreprocessor::CastInitialSamples(const int numSamples, 
										const int sampleType)
{	
	const long startTime = GetTime();

	// generate simple rays
	SimpleRayContainer simpleRays;
	GenerateRays(numSamples, sampleType, simpleRays);

	// generate vss rays
	VssRayContainer samples;
	CastRays(simpleRays, samples, true);
	// add to ray queue
	EnqueueRays(samples);

	//Debug << "generated " <<  numSamples << " samples in " << TimeDiff(startTime, GetTime()) * 1e-3 << " secs" << endl;

	return (int)samples.size();
}


void GvsPreprocessor::EnqueueRays(VssRayContainer &samples)
{
	// add samples to ray queue
	VssRayContainer::const_iterator vit, vit_end = samples.end();
	
	for (vit = samples.begin(); vit != vit_end; ++ vit)
	{
		HandleRay(*vit);
	}
}


int GvsPreprocessor::Pass()
{
	// reset samples
	int castSamples = 0;
	mSampleContriPerPass = 0;

	while (castSamples < mSamplesPerPass) 
	{	
		// Ray queue empty => 
		// cast a number of uniform samples to fill ray queue
		castSamples += CastInitialSamples(mInitialSamples, mSamplingType);
		castSamples += ProcessQueue();
		//cout << "\ncast " << castSamples << " samples in a processing pass" << endl;
	}

	mTotalSampleContri += mSampleContriPerPass;
	return castSamples;
}


int GvsPreprocessor::ProcessQueue()
{
	int castSamples = 0;
	++ mGvsPass;

	while (!mRayQueue.empty()) 
	{
		// handle next ray
		VssRay *ray = mRayQueue.top();
		mRayQueue.pop();

		castSamples += AdaptiveBorderSampling(*ray);
		delete ray;
	}
	
	return castSamples;
}


bool GvsPreprocessor::ComputeVisibility()
{
	cout << "Gvs Preprocessor started\n" << flush;
	const long startTime = GetTime();

	Randomize(0);
	
	mPass = 0;
	mGvsPass = 0;
	mSampleContriPerPass = 0; 
	mTotalSampleContri = 0;
	mReverseSamples = 0;
	mBorderSamples = 0;

	int castSamples = 0;

	if (!mLoadViewCells)
	{	
		/// construct the view cells from the scratch
		ConstructViewCells();
		cout << "finished view cell construction" << endl;
	}
	else if (1)
	{	
		//-- load view cells from file
		//-- test successful view cells loading by exporting them again
		VssRayContainer dummies;
		mViewCellsManager->Visualize(mObjects, dummies);
		mViewCellsManager->ExportViewCells("test.xml.gz", mViewCellsManager->GetExportPvs(), mObjects);
	}

	while (castSamples < mTotalSamples) 
	{
		castSamples += Pass();
				
		////////
		//-- stats
		cout << "\nPass " << mPass << " #samples: " << castSamples << " of " << mTotalSamples << endl;

		//mVssRays.PrintStatistics(mStats);
		mStats 
			<< "#Pass\n" << mPass << endl
			<< "#Time\n" << TimeDiff(startTime, GetTime())*1e-3 << endl
			<< "#TotalSamples\n" << castSamples << endl
			<< "#ScDiff\n" << mSampleContriPerPass << endl
			<< "#SamplesContri\n" << mTotalSampleContri << endl
			<< "#ReverseSamples\n" << mReverseSamples << endl
			<< "#BorderSamples\n" << mBorderSamples << endl			
			<< "#GvsRuns\n" << mGvsPass << endl;


		mViewCellsManager->PrintPvsStatistics(mStats);

		char str[64]; sprintf(str, "tmp/pass%04d-", mPass);
		
		// visualization
		const bool exportRays = true;
		const bool exportPvs = true;
		mViewCellsManager->ExportSingleViewCells(mObjects, 10, false, exportPvs, exportRays, 100, str);
		// remove pass samples
		ViewCellContainer::const_iterator vit, vit_end = mViewCellsManager->GetViewCells().end();

		for (vit = mViewCellsManager->GetViewCells().begin(); vit != vit_end; ++ vit)
		{
			CLEAR_CONTAINER((*vit)->mVssRays);
		}

		CLEAR_CONTAINER(mVssRays);
		// ComputeRenderError();
		++ mPass;
	}

	cout << 2 * castSamples / (1e3f * TimeDiff(startTime, GetTime())) << "M rays/s" << endl;
	Visualize();

	return true;
}


void GvsPreprocessor::Visualize()
{
	Exporter *exporter = Exporter::GetExporter("gvs.wrl");

	if (!exporter)
		return;
	
	vector<VizStruct>::const_iterator vit, vit_end = vizContainer.end();
	for (vit = vizContainer.begin(); vit != vit_end; ++ vit)
	{
		exporter->SetWireframe();
		exporter->ExportPolygon((*vit).enlargedTriangle);
		//Material m;
		exporter->SetFilled();
		Polygon3 poly = Polygon3((*vit).originalTriangle);
		exporter->ExportPolygon(&poly);
	}

	exporter->ExportRays(mVssRays);
	delete exporter;
}

}